/*
 * DOTmapper.js
 * by Joseph Malloch
 * copyright © 2010, IDMIL
 */

var debugFlag = 0;

var database = new SQLite;
var resultSet = new SQLResult;
var dummy = new SQLResult;
var sourceTable = new SQLResult;
var destinationTable = new SQLResult;
var linkTable = new SQLResult;
var mode = 0;
var s_order = ['device','ASC'];
var d_order = ['device','ASC'];
var s_pattern = '%';
var d_pattern = '%';
var s_scroll = 0;
var d_scroll = 0;
var s_view = new Array;
var d_view = new Array;
var s_select = new Array;
var d_select = new Array;
var l_select = new Array;
var l_click = new Array;
var link_indices = new Array;
var selectedLinks = new Array;
var destDeviceMenuSpacing = [150,75,80,50,200];
var destParameterMenuSpacing = [260,50,50,75,200];
var sourceDeviceMenuSpacing = [150,75,80,50,200];
var sourceParameterMenuSpacing = [260,50,50,75,200];

var debugOut = this.patcher.getnamed('debugOut');

var loadButton = this.patcher.getnamed('loadButton');
var saveButton = this.patcher.getnamed('saveButton');
var buttons = this.patcher.getnamed('scaling_buttons');
var expressionBox = this.patcher.getnamed('expression');
var ranges = this.patcher.getnamed('ranges');
var clipMinBox = this.patcher.getnamed('clipMin');
var clipMaxBox = this.patcher.getnamed('clipMax');

var tabs = this.patcher.getnamed('tabs');
var linkPane = this.patcher.getnamed('linkPane');
var linkStatusBar = this.patcher.getnamed('linkStatusBar');

var source = this.patcher.getnamed('source');
var sourceHeader = this.patcher.getnamed('sourceHeader');
var sourceHeaderSpacing = this.patcher.getnamed('sourceHeaderSpacing');
var sourceStatusBar = this.patcher.getnamed('sourceStatusBar');
var sourceProgressBar = this.patcher.getnamed('sourceProgressBar');
var sourceSearchBar = this.patcher.getnamed('sourceSearchBar');

var destination = this.patcher.getnamed('destination');
var destHeader = this.patcher.getnamed('destHeader');
var destHeaderSpacing = this.patcher.getnamed('destHeaderSpacing');
var destinationStatusBar = this.patcher.getnamed('destinationStatusBar');
var destinationProgressBar = this.patcher.getnamed('destinationProgressBar');
var destinationSearchBar = this.patcher.getnamed('destinationSearchBar');

var num_source = 0;
var num_dest = 0;

function loadbang() {
    clearAllDisplays();
    createTables();
    select_tab('All Devices');
    outlet(0, '/who');
    outlet(0, '/*/links/get');
}

function clearAllDisplays() {
    destination.message("clear", "all");
    destination.message("cols", 0);
    destination.message("rows", 0);
    source.message("clear", "all");
    source.message("cols", 0);
    source.message("rows", 0);
    tabs.message("clear", "all");
    tabs.message("cols", 1);
    tabs.message("rows", 0);
    tabs.message("set", 0, 0, "All Devices");
}

function createTables() {
    opendb('./dotmapper_db');
    database.exec("CREATE TABLE devices (device VARCHAR(256) UNIQUE, numInputs VARCHAR(256) DEFAULT 'unknown', numOutputs VARCHAR(256) DEFAULT 'unknown', port INT(10) DEFAULT 'unknown', IP VARCHAR(15) DEFAULT 'unknown', canAlias VARCHAR(3) DEFAULT '0', hash VARCHAR(256) DEFAULT 'unknown', receivedInputs INT(10), receivedOutputs INT(10))", dummy);
    database.exec("CREATE TABLE links (source VARCHAR(256), destination VARCHAR(256))", dummy);
    database.exec("CREATE TABLE connections (source VARCHAR(256), destination VARCHAR(256), scaling VARCHAR(256), range VARCHAR(100), expression VARCHAR(256) DEFAULT 'y=x', clipMin VARCHAR(10), clipMax VARCHAR(10))", dummy);
    database.exec("CREATE TABLE source_parameters (parameter VARCHAR(256) UNIQUE, device VARCHAR(256), type VARCHAR(25) DEFAULT ' ', units VARCHAR(100) DEFAULT ' ', minimum VARCHAR(25) DEFAULT ' ', maximum VARCHAR(25) DEFAULT ' ')", dummy);
    database.exec("CREATE TABLE dest_parameters (parameter VARCHAR(256) UNIQUE, device VARCHAR(256), type VARCHAR(25) DEFAULT ' ', units VARCHAR(100) DEFAULT ' ', minimum VARCHAR(25) DEFAULT ' ', maximum VARCHAR(25) DEFAULT ' ')", dummy);
}

function opendb(x) {  // move to loadbang
    database.open(x,0); //open a ram-based DB
}

function closedb() {  // move to closebang
	database.close();
}

function source_click() {
    if(debugFlag == 1) {
		debugOut.message('source_click()');
    }
    if((arguments.length == 2) && (sourceTable.numrecords())) {
        arguments[1] = arguments[1] / 17 + s_scroll;
        s_select[0] = sourceTable.value(0,Math.max(0,Math.min(Math.floor(arguments[1]),sourceTable.numrecords()-1)));
        s_select[1] = Math.max(0,Math.min(Math.floor(arguments[1]),sourceTable.numrecords()-1));
        //if shift held down, s_select.push[arguments[0]) instead?
        drawSource();
    }
}

function source_drag() {
    if((arguments.length == 2) && (sourceTable.numrecords()) && (destinationTable.numrecords())) {
        linkPane.message('clear');
        drawLinks();
        if(arguments[0] > 180) {
            arguments[0] = 240;
            var dest_y = Math.max(0,Math.min(Math.floor(arguments[1] / 17 + d_scroll),destinationTable.numrecords()-1));
            arguments[1] = (dest_y - d_scroll) * 17 + 9;
            d_select[0] = destinationTable.value(0,dest_y);
            d_select[1] = dest_y;
            drawDestination();
        } else {
            d_select.length = 0;
            drawDestination();
        }
        drawLine(0, (s_select[1] - s_scroll) * 17 + 9, arguments[0], arguments[1], 255, 0, 0);
    }
}

function source_doubleclick() {
    if(debugFlag == 1) {
		debugOut.message('source_doubleclick()');
    }
    if(arguments.length) {
        select_tab(arguments[1]);
    }
}

function dest_click() {
    if(debugFlag == 1) {
		debugOut.message('dest_click()');
    }
    if((arguments.length == 2) && (destinationTable.numrecords())) {
        arguments[1] = arguments[1] / 17 + d_scroll;
        d_select[0] = destinationTable.value(0,Math.max(0,Math.min(Math.floor(arguments[1]),destinationTable.numrecords()-1)));
        d_select[1] = Math.max(0,Math.min(Math.floor(arguments[1]),destinationTable.numrecords()-1));
        //if shift held down, s_select.push[arguments[0]) instead
        drawDestination();
    }
}

function dest_drag() {
    if((arguments.length == 2) && (sourceTable.numrecords()) && (destinationTable.numrecords())) {
        linkPane.message('clear');
        drawLinks();
        if(arguments[0] < 60) {
            arguments[0] = 0;
            var source_y = Math.max(0,Math.min(Math.floor(arguments[1] / 17 + s_scroll),sourceTable.numrecords()-1));
            arguments[1] = (source_y - s_scroll) * 17 + 9;
            s_select[0] = sourceTable.value(0,source_y);
            s_select[1] = source_y;
            drawSource();
        } else {
            s_select.length = 0;
            drawSource();
        }
        drawLine(arguments[0], arguments[1], 240, (d_select[1] - d_scroll) * 17 + 9, 255, 0, 0);
    }
}

function link_click() {
    if((arguments.length == 2) && (linkTable.numrecords())) {
        s_select.length = 0;
        drawSource();
        d_select.length = 0;
        drawDestination();
        l_select.length = 0;
        drawLinks();
        drawEditBar();
        l_click[0] = arguments[0];
        l_click[1] = arguments[1];
        //add click to select
    }
}

function link_drag() {
    if(arguments.length == 2) {
        l_select.length = 0;
        var xs = [l_click[0], arguments[0]];
        var ys = [l_click[1], arguments[1]];
        xs.sort(sortNumber);
        ys.sort(sortNumber);
        
        var temp1 = [xs[0],xs[1]];
        xs = [Math.cos(xs[0] / 240 * Math.PI), Math.cos(xs[1] / 240 * Math.PI)];
        var temp2 = [0, 0, ys[0], ys[1]];

        for(var i=0; i<linkTable.numrecords(); i++) {
            var temp3 = (link_indices[i*2] - link_indices[i*2+1]) * 0.5 + link_indices[i*2] - (link_indices[i*2] - link_indices[i*2+1]) * 0.5;
            temp2[0] = xs[0] * (link_indices[i*2] - link_indices[i*2+1]) * 0.5 + link_indices[i*2] - (link_indices[i*2] - link_indices[i*2+1]) * 0.5;
            temp2[1] = xs[1] * (link_indices[i*2] - link_indices[i*2+1]) * 0.5 + link_indices[i*2] - (link_indices[i*2] - link_indices[i*2+1]) * 0.5;
            
            var intersects = 0;

            if((temp2[0]<temp2[2])&&(temp2[1]>temp2[3])) {          // line enters top and exits bottom
                intersects = 1;
            } else if((temp2[1]<temp2[2])&&(temp2[0]>temp2[3])) {   // line enters bottom and exits top
                intersects = 1;
            } else if((temp2[0]>temp2[2])&&(temp2[0]<temp2[3])) {   // line passes through left side
                intersects = 1;
            } else if((temp2[1]>temp2[2])&&(temp2[1]<temp2[3])) {   // line passes through right side
                intersects = 1;
            }
                
            if(intersects) {
                l_select.push(1);
            } else
                l_select.push(0);
        }
        drawEditBar();
        drawLinks();
    }
}

function selectAll() {
    l_select.length = 0;
    for(var i = 0; i < linkTable.numrecords(); i++)
        l_select.push(1);
    drawEditBar();
    drawLinks();
}

function deleteLink() {
    if(mode == 0)
        var temp = '/unlink';
    else
        var temp = '/disconnect';
    for(var i=0; i<linkTable.numrecords(); i++) {
        if(l_select[i]) {
            outlet(0,temp,linkTable.value(0,i),linkTable.value(1,i));
        }
    }
    l_select.length = 0;
    drawSource();
    drawDestination();
}

function drawEditBar() {
    buttons.message('clear');
    ranges.message('clear');
    clipMinBox.message('clear');
    clipMaxBox.message('clear');
    if(mode == 1) {
        var scaling_temp = 'null';
        //var ranges_temp = ['null','null','null','null'];
        var expression_temp = new String('null');
        var ranges_temp = ['null', 'null', 'null', 'null'];
        var clipMin_temp = 'null';
        var clipMax_temp = 'null';
        for(var i=0; i<linkTable.numrecords(); i++) {
            if(l_select[i]) {
                if(scaling_temp == 'null')
                    scaling_temp = linkTable.value(2,i)
                else if(scaling_temp != linkTable.value(2,i))
                    scaling_temp = 'multiple';
                if(expression_temp == 'null')
                    expression_temp = linkTable.value(4,i)
                else if(expression_temp != linkTable.value(4,i))
                    expression_temp = 'y=';
                var temp = linkTable.value(3,i);
                temp = temp.split(" ");
                for(var j=0; j<4; j++) {
                    if(ranges_temp[j] == 'null')
                        ranges_temp[j] = temp[j];
                    else if(ranges_temp[j] != temp[j])
                        ranges_temp[j] = 'multiple';
                }
                if(clipMin_temp == 'null')
                    clipMin_temp = linkTable.value(5,i)
                else if(clipMin_temp != linkTable.value(5,i))
                    clipMin_temp = 'multiple';
                if(clipMax_temp == 'null')
                    clipMax_temp = linkTable.value(6,i)
                else if(clipMax_temp != linkTable.value(6,i))
                    clipMax_temp = 'multiple';
            }
        }
        if(expression_temp == 'null')
            expression_temp = 'y=';
        temp = expression_temp.indexOf('%2C');
        if(temp != -1)
            expression_temp = expression_temp.substring(0,temp) + ',' + expression_temp.substring(temp + 3, expression_temp.length);
        expressionBox.message('set',expression_temp);
        ranges.message(ranges_temp);
        clipMinBox.message(clipMin_temp);
        clipMaxBox.message(clipMax_temp);
        buttons.message(scaling_temp);
    }
}

function scaling() {
    if(mode == 1 && arguments.length) {
        for(var i=0; i<linkTable.numrecords(); i++) {
            if(l_select[i]) {
                outlet(0,'/connection/modify',linkTable.value(0,i),linkTable.value(1,i),'@scaling',arguments[0]);
            }
        }
    }
}

function expression() {
    if(mode == 1 && arguments.length) {
        for(var i=0; i<linkTable.numrecords(); i++) {
            if(l_select[i]) {
                outlet(0,'/connection/modify',linkTable.value(0,i),linkTable.value(1,i),'@scaling expression @expression',arguments[0]);
            }
        }
    }
}

function range() {
    if(mode == 1 && arguments.length == 2) {
        if(arguments[0] == 'invert') {
            for(var i=0; i<linkTable.numrecords(); i++) {
                if(l_select[i]) {
                    outlet(0,'/connection/modify',linkTable.value(0,i),linkTable.value(1,i),'@range','invert',arguments[1]);
                }
            }
        }
        else {
            var temp = ['-','-','-','-'];
            temp[arguments[0]] = arguments[1];
            for(var i=0; i<linkTable.numrecords(); i++) {
                if(l_select[i]) {
                    outlet(0,'/connection/modify',linkTable.value(0,i),linkTable.value(1,i),'@range',temp);
                }
            }
        }
    }
}

function clipMin() {
    if(mode == 1 && arguments.length) {
        for(var i=0; i<linkTable.numrecords(); i++) {
            if(l_select[i]) {
                outlet(0,'/connection/modify',linkTable.value(0,i),linkTable.value(1,i),'@clipMin',arguments[0]);
            }
        }
    }
}

function clipMax() {
    if(mode == 1 && arguments.length) {
        for(var i=0; i<linkTable.numrecords(); i++) {
            if(l_select[i]) {
                outlet(0,'/connection/modify',linkTable.value(0,i),linkTable.value(1,i),'@clipMax',arguments[0]);
            }
        }
    }
}

function sortNumber(a,b) {
    return a - b;
}

function mouse_up() {
    if(s_select.length && d_select.length) {
        if(mode == 0) {
            outlet(0,'/link',s_select[0],d_select[0]);
        } else {
            outlet(0,'/connect',s_select[0],d_select[0]);
        }
    }
    s_select.length = 0;
    d_select.length = 0;
    l_click.length = 0;
    
    drawSource();
    drawDestination();
}

function clear_selected() {
    s_select.length = 0;
    d_select.length = 0;
}

function source_scroll() {
    if(arguments.length) {
        s_scroll = arguments[0];
        drawLinks();
    }
}

function dest_scroll() {
    if(arguments.length) {
        d_scroll = arguments[0];
        drawLinks();
    }
}

function registered() {
    if(debugFlag == 1) {
		debugOut.message('registered()');
    }
    if(arguments.length) {
        var args = arrayfromargs(arguments);
        database.exec("SELECT numInputs,numOutputs,hash FROM devices WHERE device='"+arguments[0]+"'",resultSet);
        if(resultSet.numrecords()) {   //if record of device exists
            var properties = extractProperties(args);      
            
            if(properties[1][properties[0].indexOf('hash')] != resultSet.value(2,0)) { //if hash doesn't match
                database.exec("UPDATE devices SET hash='"+properties[1][properties[0].indexOf('hash')]+"', numInputs='"+properties[1][properties[0].indexOf('numInputs')]+"', numOutputs='"+properties[1][properties[0].indexOf('numOutputs')]+"', receivedInputs='0', receivedOutputs='0' WHERE device='"+args[0]+"'",resultSet);  //update hash
                database.exec("DELETE FROM source_parameters WHERE device='"+args[0]+"'",dummy); //delete device params
                database.exec("DELETE FROM dest_parameters WHERE device='"+args[0]+"'",dummy);
            }
            
            if(s_view.indexOf(args[0]) != -1)  {            //if currently viewing device outputs
                database.exec("SELECT parameter FROM source_parameters WHERE device='"+args[0]+"'",resultSet);
                if(properties[1][properties[0].indexOf('numOutputs')] != resultSet.numrecords()) {
                    outlet(0,args[0]+'/namespace/output/get');
                }
            }
            
            if(d_view.indexOf(args[0]) != -1) {      //if currently viewing device inputs
                database.exec("SELECT parameter FROM dest_parameters WHERE device='"+args[0]+"'",resultSet);
                if(properties[1][properties[0].indexOf('numInputs')] != resultSet.numrecords()) {
                    outlet(0,args[0]+'/namespace/input/get');
                }
            }
            
        } else {    //if record of device does not exist
            args.unshift('@device');
            var properties = extractProperties(args);
            database.exec(buildInsertMessage('devices',properties[0],properties[1]), dummy);  //add to database
            
            if(properties[1][properties[0].indexOf('numInputs')] > 0)
                updateDestination();
            if(properties[1][properties[0].indexOf('numOutputs')] > 0) {
                updateSource();
                drawTabs();
            }
        }
    }
}

function logout(device) {
    if(debugFlag == 1) {
		debugOut.message('logout()');
    }
    database.exec("DELETE FROM devices WHERE device='"+device+"'", dummy);
    database.exec("DELETE FROM parameters WHERE device='"+device+"'", dummy);
    database.exec("DELETE FROM links WHERE source='"+device+"' OR destination='"+device+"'", dummy);
    database.exec("DELETE FROM connections WHERE source LIKE '"+device+"%' OR destination LIKE '"+device+"%'", dummy);
    drawTabs();
    if(s_view.indexOf(device) != -1)
        select_tab('All Devices');
    else if(d_view.indexOf(device) != -1) {
        select_tab(s_view[0]);
    }
}

function linked() {
    if(arguments.length == 2) {
        if(debugFlag == 1) {
            debugOut.message('linked:', arguments[0], arguments[1]);
        }
        database.exec("SELECT source FROM links WHERE source='"+arguments[0]+"' AND destination='"+arguments[1]+"'", resultSet);
        if(resultSet.numrecords() == 0) {
            database.exec("INSERT INTO links ('source','destination') VALUES ('"+arguments[0]+"','"+arguments[1]+"')", dummy);
            if(mode == 0) {
                drawLinks();
            }
        }
    }
}

function unlinked() {
    if(debugFlag == 1) {
		debugOut.message('unlinked()', arguments);
    }
    if(arguments.length) {
        database.exec("DELETE FROM links WHERE source='"+arguments[0]+"' AND destination='"+arguments[1]+"'", dummy);
        database.exec("DELETE FROM connections WHERE source LIKE '"+arguments[0]+"%' AND destination LIKE '"+arguments[1]+"%'", dummy);
    }
    if(mode == 0) {
        drawLinks();
    }
}

function connected() {
    if(debugFlag == 1) {
		debugOut.message('connected()');
    }
    if(arguments.length >= 2) {
        var args = arrayfromargs(arguments);
        args.unshift('@source',args.shift(),'@destination');
        var properties = extractProperties(args);
        
        database.exec("SELECT scaling,range,expression,clipMin,clipMax FROM connections WHERE source='"+arguments[0]+"' AND destination='"+arguments[1]+"'",resultSet);
                
        if(resultSet.numrecords() == 0) {
            database.exec(buildInsertMessage('connections',properties[0],properties[1]),resultSet);
            if(mode == 1) {
                drawLinks();
            }
        } else {
            // test if properties differ
            var temp = [properties[1][properties[0].indexOf('scaling')],properties[1][properties[0].indexOf('range')],properties[1][properties[0].indexOf('expression')],properties[1][properties[0].indexOf('clipMin')],properties[1][properties[0].indexOf('clipMax')]];
            var changed = 0;
            for(i = 0; i<resultSet.numfields(); i++) {
                if(temp[i] != resultSet.value(i,0)) {
                    database.exec("UPDATE connections SET "+resultSet.fieldname(i)+"='"+temp[i]+"' WHERE source='"+arguments[0]+"' AND destination='"+arguments[1]+"'", dummy);
                    changed = 1;
                }
            }
            if(changed) {
                drawLinks();
                drawEditBar();
            }
        }
    }
}

function disconnected() {
    if(debugFlag == 1) {
		debugOut.message('disconnected()');
    }
    if(arguments.length == 2){
        database.exec("DELETE FROM connections WHERE source='"+arguments[0]+"' AND destination='"+arguments[1]+"'", dummy);
    }
    if(mode == 1) {
        drawLinks();
        drawEditBar();
    }
}

function anything() {
    var namespace = messagename.indexOf('/namespace');
    if(namespace != -1) {  // is "/<device>/namespace" message?
        if(debugFlag == 1) {
            debugOut.message('namespace()');
        }
        var device = "/" + messagename.substring(0,namespace);
        var IO = messagename.substring(namespace+11,messagename.length);
        if(IO == 'input' || IO == 'output') {
            var args = arrayfromargs(arguments);
            var parameter = device+args.shift();
            args.unshift('@device',device,'@parameter',parameter);
            
            if(IO == 'output') {
                database.exec("SELECT numOutputs, receivedOutputs FROM devices WHERE device='"+device+"'", resultSet);
            } else if(IO == 'input') {
                database.exec("SELECT numInputs, receivedInputs FROM devices WHERE device='"+device+"'", resultSet);
            }

            if(resultSet.numrecords()) {
                if((resultSet.value(0,0) * 1) > (resultSet.value(1,0) * 1)) {   //still missing parameter records
                    var received = resultSet.value(1,0) * 1 + 1;
                    database.exec("SELECT parameter WHERE parameter='"+parameter+"'", resultSet);   //check if parameter already recorded
                    if(resultSet.numrecords() == 0) {
                        var properties = extractProperties(args);
                        if(IO == 'output') {
                            database.exec(buildInsertMessage('source_parameters',properties[0],properties[1]), dummy);
                            database.exec("UPDATE devices SET receivedOutputs='"+received+"' WHERE device='"+device+"'", dummy);
                        } else if(IO == 'input') {
                            database.exec(buildInsertMessage('dest_parameters',properties[0],properties[1]), dummy);
                            database.exec("UPDATE devices SET receivedInputs='"+received+"' WHERE device='"+device+"'", dummy);
                        }
                    }
                }
            }
            
            if(mode == 1) {
                if(s_view.indexOf(device) != -1 && IO == 'output')
                    updateSource();
                if(d_view.indexOf(device) != -1 && IO == 'input')
                    updateDestination();
            }
        }
    }
}

function extractProperties(data) {
    var names = new Array();
    var values = new Array();
    var fresh = -1;
    for(var i=0; i<data.length; i++) {
        temp = data[i].toString();
        if(temp.charAt(0) == '@') {
            temp2 = temp.substring(1,temp.length);
            if(temp2 == 'min')
                temp2 = 'minimum';
            else if(temp2 == 'max')
                temp2 = 'maximum';
            names.push(temp2);
            fresh = 1;
        } else {
            if(fresh == -1) {
                //do nothing
            } else if(fresh == 1) {
                values.push(temp);
                fresh = 0;
            } else {
                var length = values.length - 1;
                values[length] = values[length].concat(' ',temp);
            }
        }
    }
    return [names,values];
}

function buildInsertMessage(table,names,values) {
    var temp = "INSERT INTO "+table+" ('"+names[0]+"'";
    for(var i=1; i<names.length; i++) {
        temp = temp.concat(",'",names[i],"'");
    }
    temp = temp.concat(") VALUES ('"+values[0]+"'");
    for(var i=1; i<values.length; i++) {
        temp = temp.concat(",'",values[i],"'");
    }
    temp = temp.concat(")");
    return temp;
}

function sourcePattern(val) {
    if(debugFlag == 1) {
		debugOut.message('source_order()');
    }
    if(s_pattern != val) {
        s_pattern = val;
        s_scroll = 0;
        updateSource();
    }
}

function destPattern(val) {
    if(debugFlag == 1) {
		debugOut.message('dest_pattern()');
    }
    if(d_pattern != val) {
        d_pattern = val;
        d_scroll = 0;
        updateDestination();
    }
}

function sourceOrder(val) {
    if(arguments.length) {
        if(debugFlag == 1) {
            debugOut.message('source_order()');
        }
        arguments[0] = sourceTable.fieldname(arguments[0]);
        if(arguments[0] == s_order[0]) {
            if(s_order[1] == 'ASC')
                s_order[1] = 'DESC';
            else if(s_order[1] == 'DESC')
                s_order[1] = 'ASC';
        } else {
            s_order[0] = arguments[0];
        }
        updateSource();
    }
}

function destOrder() {
    if(arguments.length) {
        if(debugFlag == 1) {
            debugOut.message('dest_order()');
        }
        arguments[0] = destinationTable.fieldname(arguments[0]);
        if(arguments[0] == d_order[0]) {
            if(d_order[1] == 'ASC')
                d_order[1] = 'DESC';
            else if(d_order[1] == 'DESC')
                d_order[1] = 'ASC';
        } else {
            d_order[0] = arguments[0];
        }
        updateDestination();
    }
}

function drawTabs() {
    if(debugFlag == 1) {
		debugOut.message('drawTabs()');
    }
    arg = "SELECT device from devices WHERE numOutputs > 0 ORDER BY device";
    database.exec(arg, resultSet);
    var num_results = resultSet.numrecords();
    
    tabs.message("clear", "all");
    tabs.message("cols", num_results+1);
    tabs.message("set", 0, 0, "All Devices");
    
    for(var i=0; i<num_results; i++) {
        tabs.message("set", i+1, 0, resultSet.value(0, i));
    }
}

function select_tab(val) {
    if(debugFlag == 1) {
		debugOut.message('select_tab(', val, ')');
    }
    s_select.length = 0;
    d_select.length = 0;
    l_select.length = 0;
    if(val == 'All Devices') {
        mode = 0;
        s_order[0] = 'device';
        d_order[0] = 'device';
        s_view.length = 0;
        d_view.length = 0;
        destHeaderSpacing.message(destDeviceMenuSpacing);
        sourceHeaderSpacing.message(sourceDeviceMenuSpacing);
        loadButton.message(0);
        saveButton.message(0);
        drawTabs();
    } else {
        mode = 1;
        s_order[0] = 'parameter';
        d_order[0] = 'parameter';
        s_view[0] = val;
        destHeaderSpacing.message(destParameterMenuSpacing);
        sourceHeaderSpacing.message(sourceParameterMenuSpacing);
        loadButton.message(1);
        destinationProgressBar.message('fgcolor', 0.33, 0.5, 1.0, 1.0);
        outlet(0, val+"/info/get");
        outlet(0, val+'/connections/get');

        database.exec("SELECT destination FROM links WHERE source='"+val+"'", resultSet);
        var num_results = resultSet.numrecords();
        d_view.length = 0;
        for(var i=0; i<num_results; i++) {
            d_view.push(resultSet.value(0, i));
            outlet(0, resultSet.value(0, i)+'/info/get');
        }
    }
    
    s_pattern = '%';
    d_pattern = '%';
    sourceSearchBar.message('clear');
    destinationSearchBar.message('clear');
    
    drawEditBar();

    s_scroll = 0;
    d_scroll = 0;
    updateSource();
    updateDestination();
}

function postDview() {
    debugOut.message('d-view =', d_view);
}

function updateSource() {
    if(debugFlag == 1) {
		debugOut.message('updateSource()');
    }
    if(mode==0) {
        database.exec("SELECT device,numInputs,numOutputs,port,IP FROM devices WHERE numOutputs > 0 AND device LIKE '"+s_pattern+"' ORDER BY "+s_order[0]+" "+s_order[1]+"", sourceTable);
        sourceStatusBar.message('set', destinationTable.numrecords(), 'devices');
        drawSource();
    } else {
        //check if received==numIO for all chosen devices
        var progress = [0, 0];
        for(var i in s_view) {
            database.exec("SELECT numOutputs,receivedOutputs FROM devices WHERE device='"+s_view[i]+"'", resultSet);
            progress[0] += resultSet.value(0,0);
            progress[1] += resultSet.value(1,0);
        }
        if(progress[0] == progress[1]) {
            if(s_view.length > 1) {
                var temp = "'"+s_view[0]+"'";
                for(var i=1; i<s_view.length; i++) {
                    temp = temp.concat(",'",s_view[i],"'");
                }
            } else
                var temp = "'"+s_view[0]+"'";
            database.exec("SELECT parameter,type,units,minimum,maximum FROM source_parameters WHERE device IN ("+temp+") AND parameter LIKE '"+s_pattern+"' ORDER BY "+s_order[0]+" "+s_order[1]+"", sourceTable);
            sourceProgressBar.message('fgcolor', 0, 0, 0, 0);
            sourceStatusBar.message('set', sourceTable.numrecords(), 'of', progress[0] * 1, 'parameters');
            drawSource();
        } else {
            sourceProgressBar.message(0, progress[1] / progress[0]);
            sourceStatusBar.message('set', progress[1] * 1, 'of', progress[0] * 1, 'parameters');
        }
    }
}

function drawSource() {
    if(debugFlag == 1) {
		debugOut.message('drawSource()');
    }
    
    var num_columns = sourceTable.numfields();
    num_source = sourceTable.numrecords();
    
    source.message("clear", "all");
    source.message("cols", num_columns);
    source.message("rows", num_source);
    
    sourceHeader.message('clear');
    if(mode == 0)
        var spacing = sourceDeviceMenuSpacing;
    else
        var spacing = sourceParameterMenuSpacing;
        
    var position = 0;

    for(var i=0; i<num_columns; i++){
        if(i)
            sourceHeader.message('linesegment',position,0,position,20);
        sourceHeader.message('moveto',position+5, 15);
        sourceHeader.message('write',sourceTable.fieldname(i));
        if(mode == 0) {
            position += sourceDeviceMenuSpacing[i];
            source.message('col',i,'width',sourceDeviceMenuSpacing[i]);
        } else {
            position += sourceParameterMenuSpacing[i];
            source.message('col',i,'width',sourceParameterMenuSpacing[i]);
        }
    }
    
    for(var i=0; i<num_source; i++){
        if(s_select.indexOf(sourceTable.value(0,i)) != -1) {
            source.message("row", i, "frgb", 255, 255, 255);
            source.message("row", i, "brgb", 0, 0, 255);
            
        } else {
            source.message("row", i, "brgb", 255-(i+1)%2*15, 255-(i+1)%2*10, 255);
            source.message("row", i, "frgb", 0, 0, 0);
        }
        for(var j=0; j<num_columns; j++) {
            source.message("set", j, i, sourceTable.value(j, i));
        }
    }
    drawLinks();
}

function sourceColWidth() {
    if(arguments.length == 2) {
        if(mode == 0) {
            sourceDeviceMenuSpacing[arguments[0]] = arguments[1];
            sourceHeaderSpacing.message(sourceDeviceMenuSpacing);
        } else {
            sourceParameterMenuSpacing[arguments[0]] = arguments[1];
            sourceHeaderSpacing.message(sourceParameterMenuSpacing);
        }
        drawSource();
    }
}

function updateDestination() {
    if(debugFlag == 1) {
		debugOut.message('updateDestination()');
    }
    if(mode==0) {
        database.exec("SELECT device,numInputs,numOutputs,port,IP FROM devices WHERE numInputs > 0 AND device LIKE '"+d_pattern+"' ORDER BY "+d_order[0]+" "+d_order[1]+"", destinationTable);
        destinationStatusBar.message('set', destinationTable.numrecords(), 'devices');
        drawDestination();
    } else {
        //check if received==numIO for all chosen devices
        var progress = [0, 0];
        for(var i in d_view) {
            database.exec("SELECT numInputs,receivedInputs FROM devices WHERE device='"+d_view[i]+"'", resultSet);
            progress[0] += resultSet.value(0,0);
            progress[1] += resultSet.value(1,0);
        }
        if(progress[0] == progress[1]) {
            if(d_view.length > 1) {
                var temp = "'"+d_view[0]+"'";
                for(var i=1; i<d_view.length; i++) {
                    temp = temp.concat(",'",d_view[i],"'");
                }
            } else
                var temp = "'"+d_view[0]+"'";
            database.exec("SELECT parameter,type,units,minimum,maximum FROM dest_parameters WHERE device IN ("+temp+") AND parameter LIKE '"+d_pattern+"' ORDER BY "+d_order[0]+" "+d_order[1]+"", destinationTable);
            destinationProgressBar.message('fgcolor', 0, 0, 0, 0);
            destinationStatusBar.message('set', destinationTable.numrecords(), 'of', progress[0] * 1, 'parameters');
            drawDestination();
        } else {
            destinationProgressBar.message(0, progress[1] / progress[0]);
            destinationStatusBar.message('set', progress[1] * 1, 'of', progress[0] * 1, 'parameters');
        }
    }
}

function drawDestination() {
    if(debugFlag == 1) {
		debugOut.message('drawDest()');
    }
    
    var num_columns = destinationTable.numfields();
    num_dest = destinationTable.numrecords();
    
    destination.message("clear", "all");
    destination.message("cols", num_columns);
    destination.message("rows", num_dest);
    
    destHeader.message('clear');
    if(mode == 0)
        var spacing = destDeviceMenuSpacing;
    else
        var spacing = destParameterMenuSpacing;
        
    var position = 0;

    for(var i=0; i<num_columns; i++){
        if(i)
            destHeader.message('linesegment',position,0,position,20);
        destHeader.message('moveto',position+5, 15);
        destHeader.message('write',destinationTable.fieldname(i));
        if(mode == 0) {
            position += destDeviceMenuSpacing[i];
            destination.message('col',i,'width',destDeviceMenuSpacing[i]);
        } else {
            position += destParameterMenuSpacing[i];
            destination.message('col',i,'width',destParameterMenuSpacing[i]);
        }
    }
    
    for(var i=0; i<num_dest; i++){
        if(d_select.indexOf(destinationTable.value(0,i)) != -1) {
            destination.message("row", i, "frgb", 255, 255, 255);
            destination.message("row", i, "brgb", 0, 0, 255);
            
        } else {
            destination.message("row", i, "brgb", 255-(i+1)%2*15, 255-(i+1)%2*10, 255);
            destination.message("row", i, "frgb", 0, 0, 0);
        }
        for(var j=0; j<num_columns; j++) {
            destination.message("set", j, i, destinationTable.value(j, i));
        }
    }
    drawLinks();
}

function destColWidth() {
    if(arguments.length == 2) {
        if(mode == 0) {
            destDeviceMenuSpacing[arguments[0]] = arguments[1];
            destHeaderSpacing.message(destDeviceMenuSpacing);
        } else {
            destParameterMenuSpacing[arguments[0]] = arguments[1];
            destHeaderSpacing.message(destParameterMenuSpacing);
        }
        drawDestination();
    }
}

function drawLinks() {
    if(debugFlag == 1) {
		debugOut.message('drawLinks()', s_scroll, d_scroll);
    }
    linkPane.message("clear");
    
    var num_sources = sourceTable.numrecords();
    var sources = new Array(num_sources);
    for(var i=0; i<num_sources; i++) {
        sources[i] = sourceTable.value(0,i);
    }
    
    var num_destinations = destinationTable.numrecords();
    var destinations = new Array(num_destinations);
    for(var i=0; i<num_destinations; i++) {
        destinations[i] = destinationTable.value(0,i);
    }
    
    if(mode == 0) {
        database.exec("SELECT * FROM links", linkTable);
    } else {
        database.exec("SELECT * FROM connections WHERE source LIKE '"+s_view+"%'", linkTable);
        if(linkTable.numrecords())
            saveButton.message(1);
        else
            saveButton.message(0);
    }

    link_indices.length = 0;
    var counter = 0;
    
    for(var i=0; i<linkTable.numrecords(); i++) {
        source_index = sources.indexOf(linkTable.value(0,i));
        dest_index = destinations.indexOf(linkTable.value(1,i));
        if(source_index != -1 && dest_index != -1) {
            if(l_select[i]) {
                source.message("row", source_index, "frgb", 255, 255, 255);
                source.message("row", source_index, "brgb", 150, 50, 50);
                destination.message("row", dest_index, "frgb", 255, 255, 255);
                destination.message("row", dest_index, "brgb", 150, 50, 50);
            }
            drawSingleLink(source_index, dest_index, l_select[i]);
            counter++;
        }
    }
    
    if(mode == 0)
        var temp = 'links';
    else
        var temp = 'connections';
    linkStatusBar.message('set', counter, 'of', linkTable.numrecords(), temp);
}

function drawSingleLink(leftIndex, rightIndex, selected) {
    leftIndex = (leftIndex - s_scroll) * 17 + 8;
    rightIndex = (rightIndex - d_scroll) * 17 + 8;
    drawLine(0, leftIndex, 240, rightIndex, selected * 255, 0, 0);
    link_indices.push(leftIndex, rightIndex);
}

function drawLine(x1, y1, x2, y2, r, g, b) {
    linkPane.message("moveto", x1, y1);
    linkPane.message("frgb", r, g, b);
    linkPane.message("pensize", 2, 2);
    if(y1 == y2)
        linkPane.message("linesegment", x1, y1, x2, y2);
    else {
        for(var i=1; i<=50; i++) {
            linkPane.message("lineto", i*(x2-x1)/50 + x1, Math.cos(i*0.02*Math.PI)*(y1-y2)*0.5 + y1 - (y1-y2)*0.5); // magic numbers!
        }
    }
}

function openJSON() {
    if((mode == 1) && arguments.length) {
        var f = new File(arguments[0]);
        if (f.isopen) {
            var setupBuf = "";
            while (f.position < f.eof) {
                var setupBuf = setupBuf + f.readline(256);
            }
            f.close();
            post("Opened file", arguments[0], '\n');
            var input = JSON.parse(setupBuf);
            for(var i in input.mapping.connections) {
                var tempSource = new Array;
                for(var j = input.mapping.sources.length - 1; j >= 0; j--) {
                    if(input.mapping.connections[i].expression.indexOf(input.mapping.sources[j].id) != -1) {  // find source from id in expression
                        tempSource.push(j);
                        input.mapping.connections[i].expression = input.mapping.connections[i].expression.replace(input.mapping.sources[j].id, "x");  // replace id with "x"
                        break;
                    }
                }
                var tempDest = new Array;
                for(var j  = input.mapping.destinations.length - 1; j >=0; j--) {
                    if(input.mapping.connections[i].expression.indexOf(input.mapping.destinations[j].id) != -1) {  // find destination from id in expression
                        tempDest.push(j);
                        input.mapping.connections[i].expression = input.mapping.connections[i].expression.replace(input.mapping.destinations[j].id, "y");  // replace id with "y"
                        break;
                    }
                }
                
                if(tempSource.length) {
                    // if multiple identical devices present, try to map them 1:1 with viewed devices?
                    for(var j in s_view) {
                        for(var k in d_view) {
                            // send /connect message
                            outlet(0,   "/connect", 
                                        s_view[j] + "/" + input.mapping.sources[tempSource[0]].parameter, 
                                        d_view[k] + "/" + input.mapping.destinations[tempDest[0]].parameter, 
                                        "@scaling", input.mapping.connections[i].scaling, 
                                        "@range", input.mapping.connections[i].range, 
                                        "@expression", input.mapping.connections[i].expression, 
                                        "@clipMin", input.mapping.connections[i].clipMin, 
                                        "@clipMax", input.mapping.connections[i].clipMax);
                        }
                    }
                }
            } 
        } else {
            post("Could not open file.");
        }
    }
}

function saveJSON() {
    //print alert if not viewing entire namespace??
    if((mode == 1) && linkTable.numrecords() && arguments.length) {
        database.exec("SELECT * FROM connections WHERE source LIKE '"+s_view+"%'", linkTable);
        var output =    { 
                            "mapping" : {
                                "fileversion" : "dot-1",
                                "sources" : [],
                                "destinations" : [],
                                "connections" : []
                            }
                        };
        var num_columns = linkTable.numfields();
        var num_links = linkTable.numrecords();
        
        for(var i=0; i<num_links; i++){
            var s = -1;
            if(output.mapping.sources.length) {
                for(j in output.mapping.sources) {
                    if(output.mapping.sources[j].parameter == linkTable.value(0,i)) {
                        s = i;
                        break;
                    }
                }   
            }
            if(s == -1) {
                s = output.mapping.sources.length;
                var tempAddress = linkTable.value(0,i);
                output.mapping.sources[s] = {
                                                "id" : "s" + s,
                                                "device" : tempAddress.substring(1,tempAddress.indexOf("/",1)),
                                                "parameter" : tempAddress.substr(tempAddress.indexOf("/",1) + 1)
                                            };
            }
            var d = -1;
            if(output.mapping.destinations.length) {
                for(j in output.mapping.destinations) {
                    if(output.mapping.destinations[j].parameter == linkTable.value(1,i)) {
                        d = i;
                        break;
                    }
                }
            }
            if(d == -1) {
                d = output.mapping.destinations.length;
                var tempAddress = linkTable.value(1,i);
                output.mapping.destinations[d] =    {
                                                        "id" : "d" + d,
                                                        "device" : tempAddress.substring(1,tempAddress.indexOf("/",1)),
                                                        "parameter" : tempAddress.substr(tempAddress.indexOf("/",1) + 1)
                                                    };
            }
            output.mapping.connections[i] = {
                                                "scaling" : linkTable.value(2,i),
                                                "range" : linkTable.value(3,i),
                                                "expression" : linkTable.value(4,i),
                                                "clipMin" : linkTable.value(5,i),
                                                "clipMax" : linkTable.value(6,i)
                                            };
                                      
            output.mapping.connections[i].expression = output.mapping.connections[i].expression.replace(/y/gi, "d" + d);    // replace "y" with destination parameter id
            output.mapping.connections[i].expression = output.mapping.connections[i].expression.replace(/x/gi, "s" + s);    // replace "x" with source parameter id
        }

        var f = new File(arguments[0],"write","JSON"); 
        if (f.isopen) {
            f.writestring(JSON.stringify(output, null, 4)); //writes a string
            f.eof = f.position;
            f.close();
            post("Mapping saved as " + arguments[0]);
        } else {
            post("Could not create file.");
        }
    }
}

// From: http://www.json.org/json2.js
// Create a JSON object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.

if (!this.JSON) {
    this.JSON = {};
}

(function () {

    function f(n) {
        // Format integers to have at least two digits.
        return n < 10 ? '0' + n : n;
    }

    if (typeof Date.prototype.toJSON !== 'function') {

        Date.prototype.toJSON = function (key) {

            return isFinite(this.valueOf()) ?
                   this.getUTCFullYear()   + '-' +
                 f(this.getUTCMonth() + 1) + '-' +
                 f(this.getUTCDate())      + 'T' +
                 f(this.getUTCHours())     + ':' +
                 f(this.getUTCMinutes())   + ':' +
                 f(this.getUTCSeconds())   + 'Z' : null;
        };

        String.prototype.toJSON =
        Number.prototype.toJSON =
        Boolean.prototype.toJSON = function (key) {
            return this.valueOf();
        };
    }

    var cx = /[\u0000\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,
        escapable = /[\\\"\x00-\x1f\x7f-\x9f\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,
        gap,
        indent,
        meta = {    // table of character substitutions
            '\b': '\\b',
            '\t': '\\t',
            '\n': '\\n',
            '\f': '\\f',
            '\r': '\\r',
            '"' : '\\"',
            '\\': '\\\\'
        },
        rep;


    function quote(string) {

// If the string contains no control characters, no quote characters, and no
// backslash characters, then we can safely slap some quotes around it.
// Otherwise we must also replace the offending characters with safe escape
// sequences.

        escapable.lastIndex = 0;
        return escapable.test(string) ?
            '"' + string.replace(escapable, function (a) {
                var c = meta[a];
                return typeof c === 'string' ? c :
                    '\\u' + ('0000' + a.charCodeAt(0).toString(16)).slice(-4);
            }) + '"' :
            '"' + string + '"';
    }


    function str(key, holder) {

// Produce a string from holder[key].

        var i,          // The loop counter.
            k,          // The member key.
            v,          // The member value.
            length,
            mind = gap,
            partial,
            value = holder[key];

// If the value has a toJSON method, call it to obtain a replacement value.

        if (value && typeof value === 'object' &&
                typeof value.toJSON === 'function') {
            value = value.toJSON(key);
        }

// If we were called with a replacer function, then call the replacer to
// obtain a replacement value.

        if (typeof rep === 'function') {
            value = rep.call(holder, key, value);
        }

// What happens next depends on the value's type.

        switch (typeof value) {
        case 'string':
            return quote(value);

        case 'number':

// JSON numbers must be finite. Encode non-finite numbers as null.

            return isFinite(value) ? String(value) : 'null';

        case 'boolean':
        case 'null':

// If the value is a boolean or null, convert it to a string. Note:
// typeof null does not produce 'null'. The case is included here in
// the remote chance that this gets fixed someday.

            return String(value);

// If the type is 'object', we might be dealing with an object or an array or
// null.

        case 'object':

// Due to a specification blunder in ECMAScript, typeof null is 'object',
// so watch out for that case.

            if (!value) {
                return 'null';
            }

// Make an array to hold the partial results of stringifying this object value.

            gap += indent;
            partial = [];

// Is the value an array?

            if (Object.prototype.toString.apply(value) === '[object Array]') {

// The value is an array. Stringify every element. Use null as a placeholder
// for non-JSON values.

                length = value.length;
                for (i = 0; i < length; i += 1) {
                    partial[i] = str(i, value) || 'null';
                }

// Join all of the elements together, separated with commas, and wrap them in
// brackets.

                v = partial.length === 0 ? '[]' :
                    gap ? '[\n' + gap +
                            partial.join(',\n' + gap) + '\n' +
                                mind + ']' :
                          '[' + partial.join(',') + ']';
                gap = mind;
                return v;
            }

// If the replacer is an array, use it to select the members to be stringified.

            if (rep && typeof rep === 'object') {
                length = rep.length;
                for (i = 0; i < length; i += 1) {
                    k = rep[i];
                    if (typeof k === 'string') {
                        v = str(k, value);
                        if (v) {
                            partial.push(quote(k) + (gap ? ': ' : ':') + v);
                        }
                    }
                }
            } else {

// Otherwise, iterate through all of the keys in the object.

                for (k in value) {
                    if (Object.hasOwnProperty.call(value, k)) {
                        v = str(k, value);
                        if (v) {
                            partial.push(quote(k) + (gap ? ': ' : ':') + v);
                        }
                    }
                }
            }

// Join all of the member texts together, separated with commas,
// and wrap them in braces.

            v = partial.length === 0 ? '{}' :
                gap ? '{\n' + gap + partial.join(',\n' + gap) + '\n' +
                        mind + '}' : '{' + partial.join(',') + '}';
            gap = mind;
            return v;
        }
    }

// If the JSON object does not yet have a stringify method, give it one.

    if (typeof JSON.stringify !== 'function') {
        JSON.stringify = function (value, replacer, space) {

// The stringify method takes a value and an optional replacer, and an optional
// space parameter, and returns a JSON text. The replacer can be a function
// that can replace values, or an array of strings that will select the keys.
// A default replacer method can be provided. Use of the space parameter can
// produce text that is more easily readable.

            var i;
            gap = '';
            indent = '';

// If the space parameter is a number, make an indent string containing that
// many spaces.

            if (typeof space === 'number') {
                for (i = 0; i < space; i += 1) {
                    indent += ' ';
                }

// If the space parameter is a string, it will be used as the indent string.

            } else if (typeof space === 'string') {
                indent = space;
            }

// If there is a replacer, it must be a function or an array.
// Otherwise, throw an error.

            rep = replacer;
            if (replacer && typeof replacer !== 'function' &&
                    (typeof replacer !== 'object' ||
                     typeof replacer.length !== 'number')) {
                throw new Error('JSON.stringify');
            }

// Make a fake root object containing our value under the key of ''.
// Return the result of stringifying the value.

            return str('', {'': value});
        };
    }


// If the JSON object does not yet have a parse method, give it one.

    if (typeof JSON.parse !== 'function') {
        JSON.parse = function (text, reviver) {

// The parse method takes a text and an optional reviver function, and returns
// a JavaScript value if the text is a valid JSON text.

            var j;

            function walk(holder, key) {

// The walk method is used to recursively walk the resulting structure so
// that modifications can be made.

                var k, v, value = holder[key];
                if (value && typeof value === 'object') {
                    for (k in value) {
                        if (Object.hasOwnProperty.call(value, k)) {
                            v = walk(value, k);
                            if (v !== undefined) {
                                value[k] = v;
                            } else {
                                delete value[k];
                            }
                        }
                    }
                }
                return reviver.call(holder, key, value);
            }


// Parsing happens in four stages. In the first stage, we replace certain
// Unicode characters with escape sequences. JavaScript handles many characters
// incorrectly, either silently deleting them, or treating them as line endings.

            text = String(text);
            cx.lastIndex = 0;
            if (cx.test(text)) {
                text = text.replace(cx, function (a) {
                    return '\\u' +
                        ('0000' + a.charCodeAt(0).toString(16)).slice(-4);
                });
            }

// In the second stage, we run the text against regular expressions that look
// for non-JSON patterns. We are especially concerned with '()' and 'new'
// because they can cause invocation, and '=' because it can cause mutation.
// But just to be safe, we want to reject all unexpected forms.

// We split the second stage into 4 regexp operations in order to work around
// crippling inefficiencies in IE's and Safari's regexp engines. First we
// replace the JSON backslash pairs with '@' (a non-JSON character). Second, we
// replace all simple value tokens with ']' characters. Third, we delete all
// open brackets that follow a colon or comma or that begin the text. Finally,
// we look to see that the remaining characters are only whitespace or ']' or
// ',' or ':' or '{' or '}'. If that is so, then the text is safe for eval.

            if (/^[\],:{}\s]*$/.
test(text.replace(/\\(?:["\\\/bfnrt]|u[0-9a-fA-F]{4})/g, '@').
replace(/"[^"\\\n\r]*"|true|false|null|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?/g, ']').
replace(/(?:^|:|,)(?:\s*\[)+/g, ''))) {

// In the third stage we use the eval function to compile the text into a
// JavaScript structure. The '{' operator is subject to a syntactic ambiguity
// in JavaScript: it can begin a block or an object literal. We wrap the text
// in parens to eliminate the ambiguity.

                j = eval('(' + text + ')');

// In the optional fourth stage, we recursively walk the new structure, passing
// each name/value pair to a reviver function for possible transformation.

                return typeof reviver === 'function' ?
                    walk({'': j}, '') : j;
            }

// If the text is not JSON parseable, then a SyntaxError is thrown.

            throw new SyntaxError('JSON.parse');
        };
    }
}());